import io
import re
from zipfile import ZipFile

from biolib.validators import validator_utils  # type: ignore


def validate_app_version(yaml_data, yaml_version, zip_file):
    files = [filename.strip('/') for filename in zip_file.namelist()]

    error_dict = {}
    validate_unsupported_root_level_fields(yaml_data, error_dict, yaml_version)
    validate_output_type(yaml_data, error_dict)
    validate_consumes_stdin(yaml_data, error_dict)
    validate_remote_hosts(yaml_data, error_dict)
    validate_description_file(yaml_data, files, zip_file, error_dict)
    validate_license_file(yaml_data, files, zip_file, error_dict)
    return error_dict


def validate_output_type(yaml_data, error_dict):
    # output_type is not required
    if 'output_type' in yaml_data.keys():
        output_type = yaml_data['output_type']
        stdout_render_types_choices = [type_tuple[0] for type_tuple in validator_utils.stdout_render_types]
        if output_type not in stdout_render_types_choices:
            error_dict['output_type'] = [
                f'Invalid output_type specified for your app. output_type can be one of {stdout_render_types_choices}'
            ]


def validate_consumes_stdin(yaml_data, error_dict):
    if 'consumes_stdin' in yaml_data.keys():
        consumes_stdin = yaml_data['consumes_stdin']
        if not isinstance(consumes_stdin, bool):
            error_dict['consumes_stdin'] = [
                f'Invalid consumes_stdin specified for your app. consumes_stdin can be true or false'
            ]


def validate_remote_hosts(yaml_data, error_dict):
    if 'remote_hosts' in yaml_data.keys():
        for hostname in yaml_data['remote_hosts']:
            # No error message is returned if the hostname is valid
            hostname_error_message = validator_utils.validate_hostname_and_return_error_message(hostname)
            if hostname_error_message:
                error_dict['remote_hosts'] = hostname_error_message


def validate_description_file(yaml_data, files, zip_file: ZipFile, error_dict):
    # Only check existence of description file if user specified it
    if 'description_file' in yaml_data.keys():
        description_path = yaml_data['description_file']

        if description_path not in files:
            error_dict['description_file'] = [
                f'Could not find description file at {description_path}. \
Please provide a path pointing to a markdown (.md) file'
            ]
        return

    else:
        description_path = 'README.md'

    if description_path in files:
        # TODO: Discuss this limit, seems very high :sweat:
        # Check if description file is bigger than 100MB, written as 100000000 bytes
        if zip_file.getinfo(description_path).file_size > 100000000:
            error_dict['description_file'] = [f'The description file {description_path} must be less than 100MB']

        validate_description_images(zip_file.open(description_path), files, zip_file, error_dict)


def validate_license_file(yaml_data, files, zip_file, error_dict):
    if 'license_file' in yaml_data.keys():
        license_path = yaml_data['license_file']

        if license_path not in files:
            error_dict['license_file'] = [
                f'Could not find license file at {license_path}. Please provide a path pointing to a license file'
            ]
            return

    else:
        license_path = 'LICENSE'

    if license_path in files:
        # Check if license file is bigger than 100MB, written as 100000000 bytes
        if zip_file.getinfo(license_path).file_size > 100000000:
            error_dict['license_file'] = [f'The license file {license_path} must be less than 100MB']


def validate_is_open_source(yaml_data, error_dict):
    if 'is_open_source' in yaml_data.keys():
        if not isinstance(yaml_data['is_open_source'], bool):
            error_dict['is_open_source'] = [
                f'Invalid is_open_source specified for your app. is_open_source can be true or false'
            ]


supported_root_level_fields_base = [
    'arguments',
    'biolib_version',
    'citation',
    'consumes_stdin',
    'description_file',
    'license_file',
    'modules',
    'output_type',
    'remote_hosts',
]

supported_root_level_fields_v1 = [
    'client_side_include',
]

supported_root_level_fields_v2 = [
    'source_files_ignore',
]


def validate_unsupported_root_level_fields(yaml_data, error_dict, yaml_version):
    if yaml_version == 1:
        supported_fields = supported_root_level_fields_base + supported_root_level_fields_v1
    else:
        supported_fields = supported_root_level_fields_base + supported_root_level_fields_v2

    errors = []
    for field in yaml_data.keys():
        if field not in supported_fields:
            errors.append(
                f'The field {field} is not valid'
            )

    if errors:
        error_dict['unsupported_fields'] = errors


def validate_description_images(description_path, files, zip_file, error_dict):
    REGEX_MARKDOWN_INLINE_IMAGE = re.compile(r'!\[(?P<alt>.*)\]\((?P<src>.*)\)')
    image_filesize_limit_in_bytes = 5000000
    supported_image_file_types = ('png', 'gif', 'jpg', 'jpeg')

    description_images = {}
    description_markdown = io.TextIOWrapper(description_path).read()

    for img_alt, img_src_path in re.findall(REGEX_MARKDOWN_INLINE_IMAGE, description_markdown):

        if img_src_path in description_images:
            continue

        if re.match(r'data:.*;base64,', img_src_path):
            error_dict['description_file'] = [
                'The Markdown description does not support base64 images. '
                'Please specify images using their path in the application files: '
                '![Example Alt Text](path/to/image.png)'
            ]
            return error_dict

        if img_src_path not in files:
            if len(img_src_path) > 200:
                img_src_path = img_src_path[:200] + '...'
            error_dict['description_file'] = [
                f'In the Markdown description the image path {img_src_path} does not exist in application files'
            ]
            return error_dict

        extension = img_src_path.split('.')[-1] if '.' in img_src_path else 'png'
        if extension not in supported_image_file_types:
            error_dict['description_file'] = [
                f'In the Markdown description, the image {img_src_path} '
                f'must point to an image of the following types {supported_image_file_types}.'
            ]
            return error_dict

        # Limit image size
        if zip_file.getinfo(img_src_path).file_size > image_filesize_limit_in_bytes:
            error_dict['description_file'] = [
                f'In the Markdown description, the image {img_src_path} is over '
                f'{image_filesize_limit_in_bytes / 1000000} MB which is too large.'
            ]

        return error_dict



from biolib import utils
from biolib.biolib_api_client import BiolibApiClient, Job
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_binary_format import ModuleInput
from biolib.biolib_binary_format.utils import RemoteIndexableBuffer
from biolib.biolib_binary_format.unencrypted_module_output import UnencryptedModuleOutput
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.biolib_logging import logger_no_user_data


class JobStorage:

    @staticmethod
    def upload_module_input(job: Job, module_input_serialized: bytes) -> None:
        base_url = BiolibApiClient.get().base_url
        job_uuid = job['public_id']
        headers = {'Job-Auth-Token': job['auth_token']}

        multipart_uploader = utils.MultiPartUploader(
            start_multipart_upload_request=dict(
                requires_biolib_auth=True,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/input/start_upload/',
                headers=headers
            ),
            get_presigned_upload_url_request=dict(
                requires_biolib_auth=True,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/input/presigned_upload_url/',
                headers=headers
            ),
            complete_upload_request=dict(
                requires_biolib_auth=True,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/input/complete_upload/',
                headers=headers
            ),
        )

        multipart_uploader.upload(
            payload_iterator=utils.get_chunk_iterator_from_bytes(module_input_serialized),
            payload_size_in_bytes=len(module_input_serialized),
        )

    @staticmethod
    def upload_module_output(job_uuid: str, module_output: bytes) -> None:
        storage_module_output = UnencryptedModuleOutput.create_from_serialized_module_output(module_output)
        logger_no_user_data.debug(f'Job "{job_uuid}" uploading result to S3...')

        # Free up memory as quickly as possible
        del module_output

        base_url = BiolibApiClient.get().base_url
        config = CloudUtils.get_webserver_config()
        compute_node_auth_token = config['compute_node_info']['auth_token']  # pylint: disable=unsubscriptable-object
        headers = {'Compute-Node-Auth-Token': compute_node_auth_token}

        multipart_uploader = utils.MultiPartUploader(
            start_multipart_upload_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/results/start_upload/',
                headers=headers,
            ),
            get_presigned_upload_url_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/results/presigned_upload_url/',
                headers=headers,
            ),
            complete_upload_request=dict(
                requires_biolib_auth=False,
                url=f'{base_url}/api/jobs/{job_uuid}/storage/results/complete_upload/',
                headers=headers,
            ),
        )

        multipart_uploader.upload(
            payload_iterator=utils.get_chunk_iterator_from_bytes(storage_module_output),
            payload_size_in_bytes=len(storage_module_output),
        )

    @staticmethod
    def get_input(job: Job) -> bytes:
        presigned_download_url = BiolibJobApi.get_job_storage_download_url(job=job, storage_type='input')
        buffer = RemoteIndexableBuffer(url=presigned_download_url)
        return ModuleInput.convert_to_serialized_module_input(buffer)

    @staticmethod
    def get_result(job: Job) -> bytes:
        presigned_download_url = BiolibJobApi.get_job_storage_download_url(job=job, storage_type='results')

        buffer = RemoteIndexableBuffer(url=presigned_download_url)
        unencrypted_module_output = UnencryptedModuleOutput(buffer)
        return unencrypted_module_output.convert_to_serialized_module_output()



import requests

from biolib.biolib_api_client.auth import BearerAuth
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_errors import BioLibError


class BiolibAccountApi:

    @staticmethod
    def fetch_by_handle(account_handle):
        response = requests.get(
            f'{BiolibApiClient.get().base_url}/api/account/{account_handle}',
            auth=BearerAuth(BiolibApiClient.get().access_token),
            timeout=5
        )

        if not response.ok:
            raise BioLibError(response.content.decode())

        return response.json()



import time
from datetime import datetime, timedelta
from multiprocessing import Process

from werkzeug.exceptions import NotFound

from biolib import utils
from biolib.biolib_logging import logger_no_user_data
from biolib.compute_node.cloud_utils import CloudUtils
from biolib.typing_utils import Dict, List
from biolib.compute_node.webserver.worker_thread import WorkerThread


JOB_ID_TO_COMPUTE_STATE_DICT: Dict = {}
UNASSIGNED_COMPUTE_PROCESSES: List = []


def get_job_compute_state_or_404(job_id: str):
    compute_state = JOB_ID_TO_COMPUTE_STATE_DICT.get(job_id)
    if compute_state:
        return compute_state

    raise NotFound('Job not found')


def get_compute_state(unassigned_compute_processes):
    if len(unassigned_compute_processes) == 0:
        start_compute_process(unassigned_compute_processes)

    return unassigned_compute_processes.pop()


def start_compute_process(unassigned_compute_processes):
    compute_state = {
        'job_id': None,
        'job': None,
        'cloud_job_id': None,
        'cloud_job': None,
        'status': {
            'status_updates': [
                {
                    'progress': 10,
                    'log_message': 'Initializing'
                }
            ],
            'stdout_and_stderr_packages_b64': []
        },
        'progress': 0,
        'job_temporary_dir': '',
        'received_messages_queue': None,
        'messages_to_send_queue': None,
        'worker_process': None
    }

    WorkerThread(compute_state).start()
    unassigned_compute_processes.append(compute_state)


def validate_saved_job(saved_job):
    if 'app_version' not in saved_job['job']:
        return False

    if 'access_token' not in saved_job:
        return False

    if 'module_name' not in saved_job:
        return False

    return True


def start_auto_shutdown_timer() -> None:
    if not utils.IS_RUNNING_IN_CLOUD:
        logger_no_user_data.error('Not running in cloud so skipping auto shutdown time start')
        return

    update_auto_shutdown_time()

    timer = CloudAutoShutdownTimer()
    timer.start()
    logger_no_user_data.debug(f"Started auto shutdown timer on pid: {timer.pid}")


def update_auto_shutdown_time() -> None:
    if not utils.IS_RUNNING_IN_CLOUD:
        logger_no_user_data.error('Not running in cloud so skipping auto shutdown time update')
        return

    webserver_config = CloudUtils.get_webserver_config()
    auto_shutdown_buffer = webserver_config['shutdown_times']['auto_shutdown_time_in_seconds']  # pylint: disable=unsubscriptable-object

    if JOB_ID_TO_COMPUTE_STATE_DICT:
        highest_max_job_compute_state = max(
            JOB_ID_TO_COMPUTE_STATE_DICT.values(),
            key=lambda compute_state: compute_state['cloud_job']['max_runtime_in_seconds']  # type: ignore
        )
        highest_max_job_runtime = highest_max_job_compute_state['cloud_job']['max_runtime_in_seconds']
    else:
        highest_max_job_runtime = 0

    auto_shutdown_time = datetime.now() + timedelta(seconds=highest_max_job_runtime) + timedelta(
        seconds=auto_shutdown_buffer
    )
    auto_shutdown_time_isoformat = datetime.isoformat(auto_shutdown_time)

    with open(CloudAutoShutdownTimer.SHUTDOWN_TIMESTAMP_FILE_PATH, 'w') as auto_shutdown_time_file:
        auto_shutdown_time_file.write(auto_shutdown_time_isoformat)

    logger_no_user_data.debug(f'Extending auto shutdown timer to: {auto_shutdown_time_isoformat}')


class CloudAutoShutdownTimer(Process):
    SHUTDOWN_TIMESTAMP_FILE_PATH = '/tmp/auto_shutdown_time.timestamp'

    def _get_auto_shutdown_time(self) -> datetime:
        with open(self.SHUTDOWN_TIMESTAMP_FILE_PATH, mode='r') as file:
            return datetime.fromisoformat(file.read())

    def run(self) -> None:
        auto_shutdown_time = self._get_auto_shutdown_time()
        while datetime.now() < auto_shutdown_time:
            time.sleep(60)
            auto_shutdown_time = self._get_auto_shutdown_time()

        logger_no_user_data.debug(f'Hit auto shutdown timer since {datetime.now()} > {auto_shutdown_time}')
        CloudUtils.deregister_and_shutdown()



from docker.models.containers import Container  # type: ignore
from docker.models.networks import Network  # type: ignore

from biolib.compute_node.job_worker.large_file_system import LargeFileSystem
from biolib.compute_node.webserver.webserver_types import ComputeNodeInfo
from biolib.typing_utils import TypedDict, Callable, Optional, List, Dict
from biolib.compute_node.remote_host_proxy import RemoteHostProxy
from biolib.biolib_api_client.app_types import Module
from biolib.biolib_api_client.job_types import Job, CloudJob


class StatusUpdate(TypedDict):
    progress: int
    log_message: str


class RemoteExecuteOptions(TypedDict):
    biolib_base_url: str
    job: Job
    root_job_id: str


SendStatusUpdateType = Callable[[StatusUpdate], None]
SendSystemExceptionType = Callable[[int], None]
SendStdoutAndStderrType = Callable[[bytes], None]


class LocalExecutorOptions(TypedDict):
    access_token: str
    biolib_base_url: str
    compute_node_info: Optional[ComputeNodeInfo]
    internal_network: Optional[Network]
    job: Job
    cloud_job: Optional[CloudJob]
    large_file_systems: Dict[str, LargeFileSystem]
    module: Module
    dns_proxy: Optional[Container]
    remote_host_proxies: List[RemoteHostProxy]
    root_job_id: str
    runtime_zip_bytes: Optional[bytes]  # TODO: replace this with a module_source_serialized
    send_status_update: SendStatusUpdateType
    send_system_exception: SendSystemExceptionType
    send_stdout_and_stderr: SendStdoutAndStderrType



# pylint: disable=unsubscriptable-object

import json
import os
import time
import tempfile
import shutil
import logging
from flask import Flask, request, Response, jsonify, after_this_request
from flask_cors import CORS  # type: ignore

from biolib import utils
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_binary_format import SavedJob
from biolib.compute_node.webserver import webserver_utils
from biolib.compute_node.cloud_utils.cloud_utils import CloudUtils
from biolib.compute_node.webserver.gunicorn_flask_application import GunicornFlaskApplication
from biolib.biolib_logging import logger, TRACE, logger_no_user_data
from biolib.compute_node.webserver.webserver_utils import get_job_compute_state_or_404

app = Flask(__name__)
CORS(app)

biolib_tmp_dir = tempfile.TemporaryDirectory()


@app.route('/hello/')
def hello():
    return 'Hello'


@app.route('/v1/job/', methods=['POST'])
def save_job():
    saved_job = json.loads(request.data.decode())

    # TODO: figure out why this shallow validate method is used
    if not webserver_utils.validate_saved_job(saved_job):
        return jsonify({'job': 'Invalid job'}), 400

    job_id = saved_job['job']['public_id']
    job_temporary_dir = os.path.join(biolib_tmp_dir.name, job_id)
    os.makedirs(job_temporary_dir)
    saved_job['BASE_URL'] = BiolibApiClient.get().base_url
    saved_job['job_temporary_dir'] = job_temporary_dir

    compute_state = webserver_utils.get_compute_state(webserver_utils.UNASSIGNED_COMPUTE_PROCESSES)
    compute_state['job_id'] = job_id
    compute_state['job'] = saved_job['job']
    compute_state['job_temporary_dir'] = job_temporary_dir

    webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id] = compute_state

    if utils.IS_RUNNING_IN_CLOUD:
        config = CloudUtils.get_webserver_config()
        saved_job['compute_node_info'] = config['compute_node_info']
        compute_state['cloud_job_id'] = saved_job['cloud_job']['public_id']
        compute_state['cloud_job'] = saved_job['cloud_job']

        webserver_utils.update_auto_shutdown_time()

    saved_job_bbf_package = SavedJob().serialize(json.dumps(saved_job))
    send_package_to_compute_process(job_id, saved_job_bbf_package)

    return '', 201


@app.route('/v1/job/<job_id>/start/', methods=['POST'])
def start_local_compute(job_id):
    module_input_package = request.data
    send_package_to_compute_process(job_id, module_input_package)
    return '', 201


@app.route('/v1/job/<job_id>/', methods=['DELETE'])
def terminate_job(job_id: str) -> Response:
    compute_state = get_job_compute_state_or_404(job_id)
    # TODO: Consider BBF package
    compute_state['received_messages_queue'].put(b'CANCEL_JOB')
    return Response()


@app.route('/v1/job/<job_id>/status/')
def status(job_id):
    # TODO Implement auth token

    compute_state = get_job_compute_state_or_404(job_id)
    current_status = compute_state['status'].copy()
    response = jsonify(current_status)

    if current_status['status_updates']:
        webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['status']['status_updates'] = []

    if current_status['stdout_and_stderr_packages_b64']:
        webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['status']['stdout_and_stderr_packages_b64'] = []

    return response


@app.route('/v1/job/<job_id>/result/')
def result(job_id):
    compute_state = webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT[job_id]
    result_is_ready = compute_state['progress'] == 95
    if result_is_ready:
        logger_no_user_data.debug(f'Job "{job_id}" result is being downloaded by the user...')
        job_temporary_dir = compute_state["job_temporary_dir"]
        module_output_path = f'{job_temporary_dir}/module_output.bbf'
        # Set the progress to 99 to inform other cleanup code that result is being downloaded
        compute_state['progress'] = 99

        @after_this_request
        def remove_job_temporary_dir(response):  # pylint: disable=unused-variable
            if os.path.exists(job_temporary_dir):
                logger_no_user_data.debug(f'Removing result dir after download from path: {job_temporary_dir}')
                shutil.rmtree(job_temporary_dir)
            return response

        return Response(open(module_output_path, 'rb'))

    else:
        logger_no_user_data.error(f'Webserver: Result for job {job_id} was not found returning 404')
        return '', 404


def send_package_to_compute_process(job_id, package_bytes):
    compute_state = get_job_compute_state_or_404(job_id)
    message_queue = compute_state['messages_to_send_queue']
    message_queue.put(package_bytes)


def start_webserver(port, host):
    def worker_exit(server, worker):  # pylint: disable=unused-argument
        active_compute_states = list(
            webserver_utils.JOB_ID_TO_COMPUTE_STATE_DICT.values()) + webserver_utils.UNASSIGNED_COMPUTE_PROCESSES
        logger.debug(f'Sending terminate signal to {len(active_compute_states)} compute processes')
        if active_compute_states:
            for compute_state in active_compute_states:
                if compute_state['worker_thread']:
                    compute_state['worker_thread'].terminate()
            time.sleep(2)
        return

    def post_fork(server, worker):  # pylint: disable=unused-argument
        logger.info('Started compute node')

        if utils.IS_RUNNING_IN_CLOUD:
            logger.debug('Initializing webserver...')
            config = CloudUtils.get_webserver_config()
            utils.IS_DEV = config['is_dev']
            BiolibApiClient.initialize(config['base_url'])

            CloudUtils.initialize()
            webserver_utils.start_auto_shutdown_timer()

    if logger.level == TRACE:
        gunicorn_log_level_name = 'DEBUG'
    elif logger.level == logging.DEBUG:
        gunicorn_log_level_name = 'INFO'
    elif logger.level == logging.INFO:
        gunicorn_log_level_name = 'WARNING'
    else:
        gunicorn_log_level_name = logging.getLevelName(logger.level)

    options = {
        'bind': f'{host}:{port}',
        'workers': 1,
        'post_fork': post_fork,
        'worker_exit': worker_exit,
        'timeout': '7200',  # Reduce to 300 when frontend no longer downloads from webserver
        'graceful_timeout': 4,
        'loglevel': gunicorn_log_level_name,
    }

    GunicornFlaskApplication(app, options).run()
